import asyncio
import logging
from collections import defaultdict
from typing import Callable

import websockets

__all__ = ["TwitchWebsocket", "ParsedMessage"]


class ParsedMessage:
    def __init__(self, message: str):
        raw_components = self._get_raw_components(message)
        if not raw_components:
            return None

        self.command = self._parse_command(raw_components["raw_command"])
        self.source = self.tags = self.parameters = None

        if self.command is None:
            return None
        else:
            if raw_components["raw_tags"] is not None:
                self.tags = self._parse_tags(raw_components["raw_tags"])

            self.source = self._parse_source(raw_components["raw_source"])
            self.parameters = raw_components["raw_parameters"]

    def __str__(self) -> str:
        return str(self.__dict__)

    def _get_raw_components(self, message: str) -> dict[str, str]:
        if not message:
            return None

        raw_tags = raw_source = raw_command = raw_parameters = ""

        # Start index
        idx = 0

        # Get tags
        if message[idx] == "@":
            end_idx = message.find(" ")
            raw_tags = message[1:end_idx]
            idx = end_idx + 1

        # Get source(nick and host)
        if message[idx] == ":":
            idx += 1
            end_idx = message.find(" ", idx)
            raw_source = message[idx:end_idx]
            idx = end_idx + 1

        # Command
        end_idx = message.find(":", idx)
        if -1 == end_idx:
            end_idx = len(message)

        raw_command = message[idx:end_idx].strip()

        # Parameters
        if end_idx != len(message):
            idx = end_idx + 1
            raw_parameters = message[idx:]

        return {
            "raw_tags": raw_tags,
            "raw_source": raw_source,
            "raw_command": raw_command,
            "raw_parameters": raw_parameters,
        }

    def _parse_command(self, raw_command: str) -> dict:
        parsed_command = None
        command_parts = raw_command.split(" ")

        match command_parts[0]:
            case "JOIN" | "PART" | "NOTICE" | "CLEARCHAT" | "HOSTTARGET":
                pass
            case "PRIVMSG":
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "PING":
                parsed_command = {"command": command_parts[0]}
            case "CAP":
                """
                The parameters part of the messages contains the
                enabled capabilities.
                """
                parsed_command = {
                    "command": command_parts[0],
                    "isCapRequestEnabled": command_parts[2] == "ACK",
                }
            case "GLOBALUSERSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {"command": command_parts[0]}
            case "USERSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {"command": command_parts[0]}
            case "ROOMSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "RECONNECT":
                logging.info(
                    "The Twitch IRC server is about to terminate the connection for maintenance."
                )
                parsed_command = {"command": command_parts[0]}
            case "421":
                logging.warning(f"Unsupported IRC command: {command_parts[2]}")
                return None
            case "001":
                # Logged in (successfully authenticated)
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "002" | "003" | "004" | "353" | "366" | "372" | "375":
                """
                Ignoring all other numeric messages.
                353 tells you who else is in the chat room you're joining.
                """
                pass
            case "376":
                logging.info(f"Numeric message: {command_parts[0]}")
                return None
            case _:
                logging.warning(f"Unexpected command: {command_parts[0]}")
                return None

        return parsed_command

    def _parse_source(self, raw_source: str) -> None | dict[str, str]:
        if not raw_source:
            return None
        else:
            source_parts = raw_source.split("!")
            return {
                "nick": source_parts[0] if len(source_parts) == 2 else None,
                "host": source_parts[1] if len(source_parts) == 2 else source_parts[0],
            }

    def _parse_tags(self, raw_tags: str):
        dict_parsed_tags = {}

        if not raw_tags:
            return dict_parsed_tags

        for tag in raw_tags.split(";"):
            tag_key, tag_value = tag.split("=")

            match tag_key:
                case "badges-info":
                    """
                    Contains metadata related to the chat badges in the badges tag.
                    Currently, this tag contains metadata only for subscriber badges, to indicate the number of months the user has been a subscriber.
                    """
                    pass
                case "badges":
                    # badges=staff/1,broadcaster/1,turbo/1;
                    if tag_value:
                        badges = dict()
                        for badge_and_version in tag_value.split(","):
                            badge, version = badge_and_version.split("/")
                            badges[badge] = version
                    else:
                        badges = None

                    dict_parsed_tags[tag_key] = badges

                case "emotes":
                    """
                    emotes=25:0-4,12-16/1902:6-10
                    emotes=emotesv2_c51307f86f6241bc8cd8385efd7c7509:0-9/emotesv2_d9f1e820ca8e42bab70fc2f22dea0d5a:31-44

                    Comma-delimited list of emotes and their positions in the message.
                    Each emote is in the form, <emote ID>:<start position>-<end position>
                    """
                    if tag_value:
                        id_to_positions = defaultdict(list)

                        for emote_id_and_pos in tag_value.split("/"):
                            emote_id, positions = emote_id_and_pos.split(":")

                            for start_end in positions.split(","):
                                start, end = start_end.split("-")

                                id_to_positions[emote_id].append((start, end))

                        dict_parsed_tags[tag_key] = dict(id_to_positions)
                    else:
                        dict_parsed_tags[tag_key] = None

                case "color":
                    dict_parsed_tags["color"] = tag_value[1:]

                case "user-id":
                    dict_parsed_tags["user-id"] = tag_value

                case "display-name":
                    dict_parsed_tags["display-name"] = tag_value

                case _:
                    pass

        return dict_parsed_tags


class TwitchWebsocket:
    def __init__(self) -> None:
        self._websocket = None

    async def connect_websocket(self):
        if self._websocket is None:
            # TODO: remove ping_timeout=None and properly detect timeout, and show on the UI when it disconnects/reconnects
            self._websocket = await websockets.connect(
                "wss://irc-ws.chat.twitch.tv:443", ping_timeout=None, ping_interval=None
            )
            await self._websocket.send(
                "CAP REQ :twitch.tv/commands twitch.tv/membership twitch.tv/tags"
            )

    async def disconnect_websocket(self):
        if self._websocket:
            await self._websocket.close()
            self._websocket = None

    async def authenticate(self, token: str, user: str):
        """
        Returns parsed authentication response
        """
        if self._websocket is None:
            raise Exception("Websocket not connected")

        await self._websocket.send(f"PASS oauth:{token}")
        await self._websocket.send(f"NICK {user}")
        await self._websocket.send(f"USER {user} 8 * :{user}")

    async def join_channel(self, channel: str):
        if self._websocket is None:
            raise Exception("Websocket not connected")

        await self._websocket.send(f"JOIN #{channel}")

    async def send_message(self, channel: str, message: str):
        logging.debug(f"Sending message on channel {channel} message: {message}")

        if self._websocket is None:
            raise Exception("Websocket not connected")

        logging.debug("Acquired lock, sending message")
        await self._websocket.send(f"PRIVMSG #{channel} :{message}")

    async def listen_message(self, callback: Callable):
        if self._websocket is None:
            raise Exception("Websocket not connected")

        raw_msg = await self._websocket.recv()

        # Multiple messages can be received at once
        for message in raw_msg.split("\r\n"):
            parsed_message = ParsedMessage(message)
            if parsed_message:
                logging.debug(
                    f"Received websocket message: {message}. Parsed as: {parsed_message}"
                )

                try:
                    if parsed_message.command.get("command", None) in (
                        "PRIVMSG",
                        "USERSTATE",  # USERSTATE messages are used to get color and user-id
                        "GLOBALUSERSTATE",
                    ):
                        await callback(parsed_message)
                except AttributeError:
                    pass



import asyncio
import logging

import uvloop

from hasherino.hasherino import Hasherino


def main():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s | %(name)s | %(filename)s | %(levelname)s | %(funcName)s | %(lineno)d | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler()],  # Outputs logs to the console
    )

    websockets_logger = logging.getLogger("websockets")
    websockets_logger.setLevel(logging.INFO)

    uvloop.install()
    app = Hasherino()
    asyncio.run(app.async_run())


if __name__ == "__main__":
    main()



"""
TODO:
- twitch emote support
- 7tv emote support
- add 7tv emotes via chat
"""
import asyncio
import logging
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Any, Awaitable

import flet as ft
import uvloop

from hasherino import helix, user_auth
from hasherino.twitch_websocket import ParsedMessage, TwitchWebsocket


class AsyncKeyValueStorage(ABC):
    async def get(self, key) -> Any:
        pass

    async def set(self, key, value):
        pass

    async def remove(self, key):
        pass


class MemoryOnlyStorage(AsyncKeyValueStorage):
    def __init__(self, page: ft.Page) -> None:
        super().__init__()
        self.page = page

    async def get(self, key) -> Any:
        return self.page.session.get(key)

    async def set(self, key, value):
        self.page.session.set(key, value)

    async def remove(self, key):
        self.page.session.remove(key)


@dataclass
class Badge:
    id: str
    name: str
    url: str


@dataclass
class User:
    name: str
    badges: list[Badge] | None = None
    chat_color: str | None = None


class EmoteSource(Enum):
    TWITCH = 0
    SEVENTV = 1


@dataclass
class Emote:
    name: str
    id: str
    source: EmoteSource

    def get_url(self) -> str:
        match self.source:
            case EmoteSource.SEVENTV:
                return f"https://cdn.7tv.app/emote/{self.id}/4x.webp"
            case EmoteSource.TWITCH:
                return f"https://static-cdn.jtvnw.net/emoticons/v2/emotesv2_{self.id}/default/dark/3.0"
            case _:
                raise TypeError


@dataclass
class Message:
    user: User
    elements: list[str | Emote]
    message_type: str


class PubSub:
    def __init__(self) -> None:
        self.funcs: set[Awaitable] = set()

    async def subscribe(self, func: Awaitable):
        self.funcs.add(func)

    async def subscribe_all(self, funcs: list[Awaitable]):
        self.funcs.update(funcs)

    async def send(self, message: Any):
        for func in self.funcs:
            await func(message)


class FontSizeSubscriber(ABC):
    async def on_font_size_changed(self, new_font_size: int):
        ...


class ChatText(ft.Text, FontSizeSubscriber):
    def __init__(self, text: str, color: str, size: int, weight=""):
        super().__init__(text, size=size, weight=weight, color=color, selectable=True)

    async def on_font_size_changed(self, new_font_size: int):
        self.size = new_font_size
        await self.page.update_async()


class ChatBadge(ft.Image, FontSizeSubscriber):
    def __init__(self, src: str, height: int):
        super().__init__(src=src, height=height)

    async def on_font_size_changed(self, new_font_size: int):
        self.height = new_font_size
        await self.page.update_async()


class ChatEmote(ft.Image, FontSizeSubscriber):
    async def on_font_size_changed(self, new_font_size: int):
        self.height = new_font_size * 2
        await self.page.update_async()


class ChatMessage(ft.Row):
    def __init__(self, message: Message, page: ft.Page, font_size: int):
        super().__init__()
        self.vertical_alignment = "start"
        self.wrap = True
        self.width = page.width
        self.page = page
        self.font_size = font_size
        self.spacing = 5
        self.vertical_alignment = ft.CrossAxisAlignment.CENTER

        self.add_control_elements(message)

    def add_control_elements(self, message):
        self.controls = [
            ChatBadge(badge.url, self.font_size) for badge in message.user.badges
        ]

        self.controls.append(
            ChatText(
                f"{message.user.name}: ",
                message.user.chat_color,
                self.font_size,
                weight="bold",
            )
        )

        for element in message.elements:
            if type(element) == str:
                result = ChatText(element, ft.colors.WHITE, self.font_size)
            elif type(element) == Emote:
                result = ChatEmote(
                    src=element.get_url(),
                    height=self.font_size * 2,
                )
            else:
                raise TypeError

            self.controls.append(result)

    async def subscribe_to_font_size_change(self, pubsub: PubSub):
        await pubsub.subscribe_all(
            [
                control.on_font_size_changed
                for control in self.controls
                if isinstance(control, FontSizeSubscriber)
            ]
        )


class SettingsView(ft.View):
    def __init__(self, font_size_pubsub: PubSub, storage: AsyncKeyValueStorage):
        self.font_size_pubsub = font_size_pubsub
        self.storage = storage

    async def init(self):
        super().__init__(
            "/settings",
            [
                ft.IconButton(icon=ft.icons.ARROW_BACK, on_click=self._back_click),
                ft.Tabs(tabs=[await self._get_appearance_tab()]),
            ],
        )

    async def _get_appearance_tab(self) -> ft.Tab:
        return ft.Tab(
            text="Appearance",
            icon=ft.icons.BRUSH,
            content=ft.Column(
                controls=[
                    ft.Text(
                        "Chat font size:",
                        size=16,
                    ),
                    ft.Slider(
                        value=await self.storage.get("chat_font_size"),
                        min=10,
                        max=50,
                        divisions=40,
                        label="{value}",
                        width=500,
                        on_change_end=self._font_size_change,
                    ),
                ],
            ),
        )

    async def _back_click(self, _):
        self.page.views.pop()
        await self.page.update_async()

    async def _font_size_change(self, e):
        await self.storage.set("chat_font_size", e.control.value)
        await self.font_size_pubsub.send(e.control.value)


class AccountDialog(ft.AlertDialog):
    def __init__(self, storage: AsyncKeyValueStorage):
        # A dialog asking for a user display name
        self.join_user_name = ft.TextField(
            label="Enter your name to join the chat",
            autofocus=True,
            on_submit=self.join_chat_click,
        )
        self.storage = storage
        super().__init__(
            open=True,
            modal=True,
            title=ft.Text("Welcome!"),
            content=ft.Column([self.join_user_name], width=300, height=70, tight=True),
            actions=[
                ft.ElevatedButton(text="Join chat", on_click=self.join_chat_click)
            ],
            actions_alignment="end",
        )

    async def join_chat_click(self, _):
        if not self.join_user_name.value:
            self.join_user_name.error_text = "Name cannot be blank!"
            await self.join_user_name.update_async()
        else:
            await self.storage.set("user_name", self.join_user_name.value)
            self.page.dialog.open = False
            await self.page.update_async()


class NewMessageRow(ft.Row):
    def __init__(self, storage: AsyncKeyValueStorage, chat_message_pubsub: PubSub):
        self.storage = storage
        self.chat_message_pubsub = chat_message_pubsub

        # A new message entry form
        self.new_message = ft.TextField(
            hint_text="Write a message...",
            autofocus=True,
            shift_enter=True,
            min_lines=1,
            max_lines=5,
            filled=True,
            expand=True,
            on_submit=self.send_message_click,
            on_focus=self.new_message_focus,
        )
        super().__init__(
            [
                self.new_message,
                ft.IconButton(
                    icon=ft.icons.SEND_ROUNDED,
                    tooltip="Send message",
                    on_click=self.send_message_click,
                ),
            ]
        )

    async def new_message_focus(self, e):
        if await self.storage.get("user_name"):
            e.control.prefix = ft.Text(f"{await self.storage.get('user_name')}: ")
            await self.page.update_async()

    async def send_message_click(self, _):
        if self.new_message.value != "":
            emote_map = {
                "catFight": Emote(
                    id="643d8003f6c0390df3367b04",
                    name="catFight",
                    source=EmoteSource.SEVENTV,
                ),
                "Slapahomie": Emote(
                    id="60f22ed831ba6ae62262f234",
                    name="Slapahomie",
                    source=EmoteSource.SEVENTV,
                ),
                "hola": Emote(
                    id="9b76f5f0f02d42738d337082c0872b2c",
                    name="hola",
                    source=EmoteSource.TWITCH,
                ),
            }
            await self.chat_message_pubsub.send(
                Message(
                    User(
                        name=await self.storage.get("user_name"),
                        badges=[],
                        chat_color="#ff0000",
                    ),
                    elements=[
                        emote_map[element] if element in emote_map else element
                        for element in self.new_message.value.split(" ")
                    ],
                    message_type="chat_message",
                )
            )
            self.new_message.value = ""
            await self.new_message.focus_async()
            await self.page.update_async()


class ChatContainer(ft.Container):
    def __init__(self, storage: AsyncKeyValueStorage, font_size_pubsub: PubSub):
        self.storage = storage
        self.font_size_pubsub = font_size_pubsub
        self.chat = ft.ListView(
            expand=True,
            spacing=10,
            auto_scroll=True,
        )
        super().__init__(
            content=self.chat,
            border=ft.border.all(1, ft.colors.OUTLINE),
            border_radius=5,
            padding=10,
            expand=True,
        )

    async def on_message(self, message: Message):
        if message.message_type == "chat_message":
            m = ChatMessage(
                message, self.page, await self.storage.get("chat_font_size")
            )
            await m.subscribe_to_font_size_change(self.font_size_pubsub)
        elif message.message_type == "login_message":
            m = ft.Text(
                message.elements[0],
                italic=True,
                color=ft.colors.WHITE,
                size=await self.storage.get("chat_font_size"),
            )
        self.chat.controls.append(m)
        await self.page.update_async()


class Hasherino:
    def __init__(
        self, font_size_pubsub: PubSub, storage: AsyncKeyValueStorage, page: ft.Page
    ) -> None:
        self.font_size_pubsub = font_size_pubsub
        self.storage = storage
        self.page = page

    async def login_click(self, _):
        app_id = await self.storage.get("app_id")
        token = await user_auth.request_oauth_token(app_id)
        users = await helix.get_users(app_id, token, [])

        if users:
            await self.storage.set("token", token)
            await self.storage.set("user_name", users[0].display_name)
            await self.storage.set("user", users[0])
            socket: TwitchWebsocket = await self.storage.get("websocket")
            await self.storage.set(
                "ttv_badges", await helix.get_global_badges(app_id, token)
            )
            await socket.authenticate(token, users[0].login)
        else:
            self.page.dialog = ft.AlertDialog(
                content=ft.Text("Failed to authenticate.")
            )
            self.page.dialog.open = True

        await self.page.update_async()

    async def get_badge(self, set_id: str, version: str) -> dict | None:
        try:
            emotes = await self.storage.get("ttv_badges")
            id_match = next((s for s in emotes if s["set_id"] == set_id))
            version_match = next(
                (s for s in id_match["versions"] if s["id"] == version)
            )
            return version_match
        except:
            return None

    async def settings_click(self, _):
        sv = SettingsView(self.font_size_pubsub, self.storage)
        await sv.init()
        self.page.views.append(sv)
        await self.page.update_async()

    async def select_chat_click(self, _):
        async def message_received(message: ParsedMessage):
            if message.command["command"] != "PRIVMSG":
                return

            author: str = message.source["nick"]
            color = message.tags["color"]
            badges: list[Badge] = []
            for id, version in message.tags["badges"].items():
                badge = await self.get_badge(id, version)
                if badge:
                    badges.append(Badge(id, badge["title"], badge["image_url_4x"]))
            message_text: str = message.parameters
            emote_map = {}

            await self.chat_message_pubsub.send(
                Message(
                    User(
                        name=author,
                        badges=badges,
                        chat_color=f"#{color}",
                    ),
                    elements=[
                        emote_map[element] if element in emote_map else element
                        for element in message_text.split(" ")
                    ],
                    message_type="chat_message",
                )
            )

        channel = ft.TextField(label="Channel")

        async def join_chat_click(_):
            websocket: TwitchWebsocket = await self.storage.get("websocket")
            self.page.dialog.open = False
            await self.page.update_async()
            await websocket.join_channel(channel.value)

            while True:
                task = asyncio.create_task(websocket.listen_message(message_received))
                while not task.done():
                    await asyncio.sleep(0.3)

        self.page.dialog = ft.AlertDialog(
            content=channel,
            actions=[ft.ElevatedButton(text="Join", on_click=join_chat_click)],
        )
        self.page.dialog.open = True
        await self.page.update_async()

    async def run(self):
        self.page.horizontal_alignment = "stretch"
        self.page.title = "hasherino"

        self.page.dialog = AccountDialog(self.storage)

        chat_container = ChatContainer(self.storage, self.font_size_pubsub)
        self.chat_message_pubsub = PubSub()
        await self.chat_message_pubsub.subscribe(chat_container.on_message)

        # Add everything to the page
        await self.page.add_async(
            ft.Row(
                [
                    ft.IconButton(icon=ft.icons.LOGIN, on_click=self.login_click),
                    ft.IconButton(icon=ft.icons.CHAT, on_click=self.select_chat_click),
                    ft.IconButton(icon=ft.icons.SETTINGS, on_click=self.settings_click),
                ]
            ),
            chat_container,
            NewMessageRow(self.storage, self.chat_message_pubsub),
        )


async def main(page: ft.Page):
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s | %(name)s | %(filename)s | %(levelname)s | %(funcName)s | %(lineno)d | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler()],  # Outputs logs to the console
    )

    websockets_logger = logging.getLogger("websockets")
    websockets_logger.setLevel(logging.INFO)

    storage = MemoryOnlyStorage(page)
    websocket = TwitchWebsocket()
    asyncio.gather(
        websocket.connect_websocket(),
        storage.set("chat_font_size", 18),
        storage.set("app_id", "hvmj7blkwy2gw3xf820n47i85g4sub"),
        storage.set("websocket", websocket),
    )
    hasherino = Hasherino(PubSub(), storage, page)
    await hasherino.run()


def run_hasherino():
    # Script entrypoint
    uvloop.install()
    ft.app(target=main)


if __name__ == "__main__":
    run_hasherino()



import asyncio
import logging
from random import randint
from typing import Callable
from webbrowser import open as wb_open

from aiohttp import ClientSession, web

# Field from implicit grant flow used to prevent CSRF attacks
_STATE = str(randint(1, 100_000_000))

_token = ""
_error = ""

__all__ = ["request_oauth_token"]


class _WebServerContextManager:
    def __init__(self) -> None:
        self._runner: web.AppRunner

    async def __aenter__(self):
        app = web.Application()
        app.add_routes(
            [web.get("/", _home_callback),
             web.get("/auth", _browser_redirect_callback)]
        )
        self._runner = web.AppRunner(app)
        await self._runner.setup()
        site = web.TCPSite(self._runner, "localhost", 17563)
        await site.start()

    async def __aexit__(self, exc_type, exc, tb):
        await self._runner.shutdown()
        await self._runner.cleanup()


async def _is_valid_token(token: str) -> bool:
    if not token:
        return False

    async with ClientSession() as session:
        async with session.get(
            "https://id.twitch.tv/oauth2/validate",
            headers={"Authorization": f"OAuth {token}"},
        ) as response:
            return response.status == 200


async def _browser_redirect_callback(request: web.Request) -> web.Response:
    """
    Gets state and token passed by _home_callback js code
    """
    logging.info("Running auth callback")

    global _token

    state = request.rel_url.query.get("state")
    _token = request.rel_url.query.get("access_token")

    if state != _STATE:
        _error = f"Invalid state provided: {state}"
        return web.Response(text=_error)

    if not _token:
        _error = "No token found."
        return web.Response(text=_error)

    return web.Response(text="Authenticated. You may close this tab.")


async def _home_callback(_: web.Request) -> web.Response:
    """
    Sends URL parameters passed by twitch as fragments(readable client-side only) to the auth route
    """
    with open("UserAuth.html", "r") as file:
        return web.Response(text=file.read(), content_type="text/html")


async def request_oauth_token(app_id: str, existing_token: str = "") -> str:
    """
    Validate existing token or ask user to authenticate with twitch and provide a new one.

    Raises Exception if an issue occurs while getting the token
    """
    if await _is_valid_token(existing_token):
        logging.info("Provided token is valid, returning")
        return existing_token

    headers = {
        "client_id": app_id,
        "redirect_uri": "http://localhost:17563",
        "response_type": "token",
        "scope": "chat:edit chat:read user:manage:chat_color",
        "state": _STATE,
    }
    url_formatted_headers = "&".join(
        f"{header}={value}" for header, value in headers.items()
    )

    # Open user's default web browser to request the OAuth token
    url = f"https://id.twitch.tv/oauth2/authorize?{url_formatted_headers}"
    wb_open(url)

    logging.info(f"Waiting for twitch OAuth token from URL: {url}")

    async with _WebServerContextManager():
        while not _token and not _error:
            await asyncio.sleep(1)

        if _error:
            raise Exception(_error)

        logging.info("Returning found token")
        return _token
